
#include "butterworth.h"
#include <avr/pgmspace.h>
#include <EEPROM.h>

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

#define VIN           0         // ADC channel input definitions, adopt for particular builds and conditions
#define IIN           1
#define VOUT          2
#define IOUT          3
#define INDUCTOR      6         // ADC6,7 only available in Nano, not Uno
#define HEATSINK      7

#define NIGHT   0
#define MPPT    1
#define ABSORB  2
#define FLOAT   3
#define EQ      4

#define NTIMEOUTS     7
#define SEC_MIN       60.0       // 1.0 for testing, 60.0 for reals

#define HI_LOW        0          // = 1 if the NTC thermistors are connected to 5V, = 0 if they are connected to GND

typedef struct  {
  float vin_cal;            // 4 cal factors to be applied to ADC value, converted to float before multiplying
  float vout_cal;           //
  float iin_cal;            //
  float iout_cal;           //
  int z_iin;                //
  int z_iout;               // ADC zero offsets for 2 bipolar current sensors, still maybe an offset for unipolar?
  float vbat_float;         // target for floating battery
  float vbat_absorb;        // charge voltage target for battery in absorb mode
  float hs_max;             // max temp for heatsink
  float inductor_max;       // max temp, if execeeded shut down controller
  float absorb_tol;         // if battery drops from target by this much, change back to mppt, or stop counting it as EQ
  float vbat_max;           // over voltage protection for battery
  float imax;               // max current output by controller. 45A
  float night_pwr;
  float vin_max;            // max input voltage. If exceeded shutdown and wait for input caps to discharge
  float vbat_eq;            // Equalisation voltage
  float eq_duration;        // minutes TO RUN it
  int eq_started;           // flag. =0 if not in eq mode, = 1 if in eq mode. This to enable restarting or stopping
  float eq_reset;           // if during EQ, battery drops below this for one minute, EQ is stopped and back to absorb.
  float absorb_float_time;  // time at absorb needed before transion to float
  float float_tol;          // voltage battery can drop to before float mode is reset
  int fast_oc_enable;       // enables the fast comparator based OC shutdown.
  int pwm_freq;             // 20 or 40
  int timer1_period;        // 833 for 20kHz, 416 for 40kHz, actually 19,200 and 38,400
  int baud_waits;           // = 2 for 20kHz, = 4 for 40kHz
  unsigned int ints_second; // = 19200 for "20kHz", = 38400 for "40kHz".. need unsigned int to allow 38400 which is > 32768
  float fan_on;             // heat sink temperatures for fan control
  float fan_off;
  float hs_throttle_low;    // the heatsink temp where output current is reduced linearly from hs_throttle_low to hs_max. 100% throttle at high.
} _nvdata;

typedef struct {
  float total_energy;       // running total of energy output by controller. Updated once an hour.
} _nvtotals;

String spstr = "";          // for serial console data storage
uint8_t got_data=0;

int dd=0;

volatile uint8_t blcd,go_lcd = 0,bpos=0;
volatile int baudc = 0,freq_scale;
unsigned int ints_second;
int baud_waits,timer1_period;

float max_pwr, old_pwr, old_i,old_v, max_pwr_pwm;
volatile float ipwr, opwr, vin, vout, iin, iout, hs_temp, inductor_temp, hours_energy, absorb_time,eq_time,float_time;
volatile int erno, pwm, cc, track_mode, d_pwm,  one_sec_flag, one_min,calibrate;
volatile int one_min_flag, ok_temps, is_running, mp_scan_dir,hour_count, hour_flag;
volatile int one_sec,dlay_timer;   // with pwm at 40kHz, the one second timer count will overflow signed ints

int sample_count,nsamples,old_pwm,pause_data,eq_button, fan_on_flag,hs_throttle_flag;
volatile float tm_pwr,bv_target,hs_throttle_current;

float slow_iout,slow_vout,slow_iin,slow_vin;
volatile int timeout_count[NTIMEOUTS],timeout_max[NTIMEOUTS];

_nvtotals nvt;
_nvdata nvd;

volatile char buf[85],*bp,bcount;         // local text buffer for LCD

void show_menu();
void parse_it();
void(* reset_uc) (void) = 0; //declare reset function @ address 0

void serialEvent() 
{
  char c;
  while (Serial.available()) 
    {
    c = (char)Serial.read();
    if (got_data == 0)    // dont store while still working on previous string
      {
      spstr += c;
      if (c == '\n')
        got_data = 1;
      }
    }
}

void show_menu()
  {  
  get_nvdata();
  spstr = "";
  got_data=0;
  Serial.println(F("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"));
  Serial.print(F("A - max heat sink temp   "));Serial.print(nvd.hs_max,0);Serial.println(F(" degC "));
  Serial.print(F("B - max inductor temp    "));Serial.print(nvd.inductor_max,0);Serial.println(F(" degC "));
  Serial.print(F("4 - H.S. throttle low    "));Serial.print(nvd.hs_throttle_low,0);Serial.println(F(" degC "));
  Serial.print(F("1 - Fan ON temp          "));Serial.print(nvd.fan_on,0);Serial.println(F(" degC "));
  Serial.print(F("2 - Fan OFF temp         "));Serial.print(nvd.fan_off,0);Serial.println(F(" degC "));
  Serial.println(" ");
  Serial.print(F("C - cal. Input Volts     ")); Serial.print(nvd.vin_cal,4);Serial.print(" ( ");Serial.print(vin);Serial.println(" V) ");
  Serial.print(F("D - cal. Output Volts    ")); Serial.print(nvd.vout_cal,4);Serial.print(" ( ");Serial.print(vout);Serial.println(" V) ");
  Serial.print(F("E - cal. Input Amps      ")); Serial.print(nvd.iin_cal,4);Serial.print(" ( ");Serial.print(iin);Serial.println(" A) ");
  Serial.print(F("F - cal. Output Amps     ")); Serial.print(nvd.iout_cal,4);Serial.print(" ( ");Serial.print(iout);Serial.println(" A) ");
  Serial.print(F("G - zero Iin sensor      ")); Serial.print(nvd.z_iin);Serial.println(F(" ADC counts"));
  Serial.print(F("H - zero Iout sensor     ")); Serial.print(nvd.z_iout);Serial.println(F(" ADC counts"));
  Serial.println(" ");
  Serial.print(F("I - Max battery Volts    ")); Serial.print(nvd.vbat_max);Serial.println(F(" Volts"));
  Serial.print(F("J - Absorb voltage       ")); Serial.print(nvd.vbat_absorb);Serial.println(F(" Volts"));
  Serial.print(F("K - Absorb tolerence     ")); Serial.print(nvd.absorb_tol);Serial.println(F(" Volts"));
  Serial.println(" ");
  Serial.print(F("L - Maximum Current      ")); Serial.print(nvd.imax);Serial.println(F(" Amps"));
  Serial.print(F("M - Night power          ")); Serial.print(nvd.night_pwr);Serial.println(F(" Watts"));
  Serial.print(F("N - Zero Energy total    ")); Serial.print(nvt.total_energy);Serial.println(F(" kW.hr"));
  Serial.println(" ");
  Serial.print(F("O - Equalisation voltage ")); Serial.print(nvd.vbat_eq);Serial.println(F(" Volts"));
  Serial.print(F("P - EQ duration          ")); Serial.print(nvd.eq_duration);Serial.println(F(" minutes"));
  Serial.print(F("Q - EQ reset tolerance   ")); Serial.print(nvd.eq_reset);Serial.println(F(" Volts"));
  Serial.println(" ");
  Serial.print(F("R - Absorb duration      ")); Serial.print(nvd.absorb_float_time);Serial.println(F(" minutes"));
  Serial.print(F("S - Float voltage        ")); Serial.print(nvd.vbat_float);Serial.println(F(" Volts"));
  Serial.print(F("T - Float tolerence      ")); Serial.print(nvd.float_tol);Serial.println(F(" Volts"));
  Serial.println(" ");
  Serial.print(F("U - fast OC enable       ")); Serial.print(nvd.fast_oc_enable);Serial.println(F(" (ON = 1, OFF = 0)"));
  Serial.print(F("V - pwm freq.            ")); Serial.print(nvd.pwm_freq);Serial.println(F("kHz (20 or 40)"));
  Serial.println(" ");
  Serial.print(F("Y - Calibrate mode       ")); Serial.print(calibrate);Serial.println(F(" (1 = ON, 0 = OFF)"));
  Serial.print(F("Z - Write defaults       ")); Serial.println(" ");
  }

void parse_it()
  {
  int i;
  float f,v;
  char c,*p;
  spstr.toUpperCase();
  c = spstr.charAt(0);
  switch (c)
    { 
    case '^':         // this probably used by myself only, to allow remote control of controller via the raspberry pi & USB serial
      spstr.setCharAt(0,' ');                           // remove ^, so as to allow conversion to int
      i = spstr.toInt();                                // send ^nn<newline>
      if (i == 42)        Serial.println(F("the answer to life, the universe and everything"));
      if (i == 12)        Serial.println(vin);    
      if (i == 13)        Serial.println(vout);
      if (i == 14)        Serial.println(iin);
      if (i == 15)        Serial.println(iout);
      if (i == 16)        Serial.println(hs_temp);
      if (i == 17)        Serial.println(inductor_temp);
      if (i == 18)        Serial.println(track_mode);
      if (i == 19)        Serial.println(pwm);
      if (i == 20)        Serial.println(nvt.total_energy + hours_energy);      
      if (i == 1 )        Serial.println(F("mppt controller v.1"));
      break;      
    case '?':
      show_menu();
      break;  
    case '-':
      if( pause_data == 0)
        pause_data = 1;
      else
        pause_data = 0;
      break;  
    case 'Z': // defaults
      nvd.vin_cal = 0.0903;
      nvd.vout_cal = 0.0903;  
      nvd.iin_cal = 0.246;  
      nvd.iout_cal = 0.246;  
      nvd.z_iin = 511;  
      nvd.z_iout = 509;  
      nvd.vbat_float =  13.0;
      nvd.vbat_absorb = 13.5;
      nvd.hs_max = 100.0;
      nvd.inductor_max = 100.0;
      nvd.absorb_tol = 0.5;
      nvd.vbat_max = 16.0;
      nvd.imax = 35.0;
      nvd.night_pwr = 10.0;
      nvd.vbat_eq = 15.0;
      nvd.eq_duration = 60.0;
      nvd.eq_reset = 0.5;
      nvd.absorb_float_time = 60.0;
      nvd.float_tol = 0.5;
      nvd.fast_oc_enable = 0;
      nvd.pwm_freq = 20;
      nvd.fan_on = 50.0;
      nvd.fan_off = 30.0;
      nvd.timer1_period = 833;
      nvd.baud_waits = 1;
      nvd.ints_second = 19200; 
      nvd.hs_throttle_low = nvd.hs_max - 10.0;
      nv();
      break;
    case 'C': // cal input volts
      Serial.println(F("Input volts cal. Enter volts "));
      f = get_float();
      i = adc_average(VIN,25000);
      nvd.vin_cal = f/(float)i;
      nv();
      break;           
    case 'D': // cal output volts
      Serial.println(F("Output volts cal. Enter volts "));
      f = get_float();
      i = adc_average(VOUT,25000);
      nvd.vout_cal = f/(float)i;
      nv();
      break;         
    case 'E': // cal input current
      Serial.println(F("Input current cal. Enter amps"));
      f = get_float();
      i = adc_average(IIN,25000);
      nvd.iin_cal = f/ ((float)(i - nvd.z_iin));
      nv();
      break;       
    case 'F': // cal output current
      Serial.println(F("Output current cal. Enter amps"));
      f = get_float();
      i = adc_average(IOUT,25000);
      nvd.iout_cal = f/ ((float)(i - nvd.z_iout));
      nv();
      break; 
    case 'A': //  Heatsink max. temp
      Serial.println(F("Heatsink maximum temperature (degC)"));
      nvd.hs_max = get_float();
      nv();
      break;            
    case '4': //  Heatsink throttle_low
      Serial.println(F("Heatsink throttle LOW temp (degC)"));
      nvd.hs_throttle_low = get_float();
      if (nvd.hs_throttle_low > (nvd.hs_max - 1.0))
        nvd.hs_throttle_low = (nvd.hs_max - 1.0);
      if (nvd.hs_throttle_low < 0.0)
        nvd.hs_throttle_low = 0.0;
      nv();
      break;    
    case 'B': //  inductor max. temp
      Serial.println(F("Inductor maximum temperature (degC)"));
      nvd.inductor_max = get_float();
      nv();
      break;   
    case '1': //  Heatsink fan
      Serial.println(F("Heatsink fan ON temperature (degC)"));
      nvd.fan_on = get_float();
      nv();
      break;
    case '2': //  Heatsink fan
      Serial.println(F("Heatsink fan OFF temperature (degC)"));
      nvd.fan_off = get_float();
      nv();
      break; 
    case 'G': //  zero input current
      i = adc_average(IIN,25000);
      nvd.z_iin = i;
      nv();
      break;    
    case 'H': //  zero output current
      i = adc_average(IOUT,25000);
      nvd.z_iout = i;
      nv();
      break;    
    case 'I': //  Max battery Volts
      Serial.println(F("Maximum battery voltage before forced shutdown "));
      nvd.vbat_max = get_float();
      nv();
      break;    
    case 'J': //  
      Serial.println(F("Absorb voltage "));
      nvd.vbat_absorb = get_float();
      nv();
      break;      
    case 'K': //  exceed this to force back to mppt from absorb. It resets the absorb timer too
      Serial.println(F("Absorb tolerence (Volts) "));
      nvd.absorb_tol = get_float();
      nv();
      break;  
    case 'L': //  output current limit
      Serial.println(F("Maximum output current "));
      nvd.imax = get_float();
      nv();
      break;          
    case 'M': //  minimum to get out of bed
      Serial.println(F("Night mode power limit (W) "));
      nvd.night_pwr = get_float();
      nv();
      break; 
    case 'N': //  zero energy total
      nvt.total_energy = 0.0;
      nv();
      break; 
    case 'O': //  EQ volts
      Serial.println(F("Equalisation voltage (Volts) "));
      nvd.vbat_eq = get_float();
      nv();
      break; 
    case 'P': //  EQ duration
      Serial.println(F("Equalisation duration (minutes) "));
      nvd.eq_duration = get_float();
      nv();
      break;     
    case 'Q': //  EQ reset
      Serial.println(F("Equalisation reset voltage (Volts) "));
      nvd.eq_reset = get_float();
      nv();
      break;   
    case 'R': //  
      Serial.println(F("Absorb duration for going to float (mminutes) "));
      nvd.absorb_float_time = get_float();
      nv();
      break;   
    case 'S': //  float volts
      Serial.println(F("Float voltage (Volts) "));
      nvd.vbat_float = get_float();
      nv();
      break;    
    case 'T': //  float tolerence
      Serial.println(F("Float tolerence (Volts) "));
      nvd.float_tol = get_float();
      nv();
      break;   
    case 'U': //  OC enable
      Serial.println(F("Enable OC shutdown (1 = ON, 0 = OFF) "));
      nvd.fast_oc_enable = get_int();
      nv();
      break;  
    case 'V': //  pwm freq
      Serial.println(F(" pwm freq.  either 20 or 40 (kHz) "));
      i = get_int();
      if (i != 20 && i != 40)
        {
          Serial.println(F(" invalid input - Setting it to 20kHz.. "));
          timer1_period = nvd.timer1_period = 833;
          baud_waits = nvd.baud_waits = 1;
          ints_second = nvd.ints_second = 19200;
          nvd.pwm_freq = 20;
          freq_scale = 1;
        }
      else
        {
          if (i == 20)
            {
            timer1_period = nvd.timer1_period = 833;
            baud_waits = nvd.baud_waits = 1;
            ints_second = nvd.ints_second = 19200; 
            nvd.pwm_freq = 20;
            freq_scale = 1;
            }
          if (i == 40)
            {
            timer1_period = nvd.timer1_period = 415;
            baud_waits = nvd.baud_waits = 3;
            ints_second = nvd.ints_second = 38400; 
            nvd.pwm_freq = 40;
            freq_scale = 2;
            }
        }
      nv();
      setup();
      break;    
    case 'Y':     //  Cal mode. This disables tracking. Use this when calibrating a controller for first time
    Serial.println(F("Enable calibrate mode (1 = ON, 0 = OFF) "));
      calibrate = get_int();
      nv();  
      break;                                                                                                          
    default:   
      spstr="";
      got_data = 0;
  }
  spstr="";
  got_data = 0;
 }

int adc_average(int ch, int n)
  {
  float tv;
  int i;
  tv = 0.0;
  for(i=0; i < n; i++)
    tv += (float)analogRead(ch);
  tv /= (float)n;
  return (int)tv;
  }

float get_float()
  {
    char c;
    got_data = 0;
    spstr = "";
    while (got_data == 0)
      {
      if (Serial.available()) 
        {
        c = (char)Serial.read();
        spstr += c;
        if (c == '\n')
          got_data = 1;
        }
      }
  return spstr.toFloat();
  }

int get_int()
  {
    char c;
    got_data = 0;
    spstr = "";
    while (got_data == 0)
      {
      if (Serial.available()) 
        {
        c = (char)Serial.read();
        spstr += c;
        if (c == '\n')
          got_data = 1;
        }
      }
  return spstr.toInt();
  }

void get_nvdata()
{
 EEPROM.get(10,nvd);
 EEPROM.get(2,nvt); 
}

void nv()
{
  EEPROM.put(10,nvd);
  EEPROM.get(10,nvd);
  EEPROM.put(2,nvt); 
  EEPROM.get(2,nvt); 
  show_menu();
}

int timeout_check(int i,int t)
  {   // this takes an index and a condition test
  if (t)
    {                         
    timeout_count[i]++;        
    if(timeout_count[i] > timeout_max[i])
      {
        timeout_count[i]=0;
        return 1;
      }
    else      
      return 0;
    }
  else
    {
    timeout_count[i]--;
      if(timeout_count[i] < 0)
        timeout_count[i] = 0; 
    return 0; 
    }
  }
  
float get_bv_track_mode()
  {
    if(track_mode == NIGHT || track_mode == MPPT || track_mode == ABSORB)
      return nvd.vbat_absorb;
    if(track_mode == FLOAT)
       return nvd.vbat_float;
    if(track_mode == EQ)
      return nvd.vbat_eq;
    return -1;
  }

void buck_stop(int v)
  {
    int i;
    erno = 3;
    //Serial.print(" stopping ");    Serial.print(v);    Serial.println();
    //ramp down to small pwm from start pwm
    for(i=pwm; i > 25; i-=20)    // should complete with pwm = 5;
      {
      pwm = i;
      dlay(800);
      }
    pwm = 5;
    is_running = 0;
    pwm = 1;
    pinMode(9, INPUT);    // high impedence output on pin, incapable of driving MOSFET gate, i.e. shutdown
  }

void buck_start()
  {
    TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(WGM11);
    erno = 4;
    pinMode(9, OUTPUT);     // enable output to TLV-250 opto, there should be a pulldown resistor across this and ground
    is_running = 1;         // to prevent Arduino boot up glitch going to TLV-250 and then to gate drive outputs.
    pwm = 1;                  
  }

void init_controller()
  {
   // init state 
  eq_button = 0; 
  track_mode = MPPT;   
  get_nvdata();      // read EEPROM data into global memory 
  bv_target = get_bv_track_mode(); 
  pwm = 1;
  hs_throttle_current = nvd.imax;
  max_pwr_pwm = 10;
  fan_on_flag = 0;
  absorb_time = 0.0;
  one_sec = 0;
  one_sec_flag = 0;
  one_min = 0;
  one_min_flag = 0;
  hour_flag = 0;
  hour_count = 0;
  old_pwr = 0.0;
  old_pwm = 0;
  sample_count = 0;
  nsamples = 150;
  tm_pwr = 0.0;
  buck_stop(0);
  ok_temps = 0;
  hours_energy = 0.0;
  slow_iout = 0.0;
  slow_vout = 0.0; 
  slow_iin = 0.0;
  slow_vin = 0.0; 
  pause_data = 0;
  calibrate = 0;
  timeout_count[0] = 0;     // vin < target timeout, in seconds
  timeout_max[0] = 30;
  timeout_count[1] = 0;     // vout > target timeout
  timeout_max[1] = 20;
  timeout_count[2] = 0;     // for when battery voltage stays lower than eq_reset voltage. Will cause cancel of EQ.
  timeout_max[2] = 30;
  timeout_count[3] = 0;     // for when battery voltage stays lower than float reset voltage. Will cancel float
  timeout_max[3] = 30;
  timeout_count[4] = 0;     // timeout for absorb time totaliser. BV must be low for 10 seconds to cancel absorb total
  timeout_max[4] = 10;
  timeout_count[5] = 0;     // timeout for over current reboot, give it a minute
  timeout_max[5] = 60;
  dd=0;
  if (nvd.pwm_freq != 20 && nvd.pwm_freq != 40)
    {
    // set default at 20 kHz
    timer1_period = 833;
    baud_waits = 1;
    ints_second = 19200;
    freq_scale = 1;
    }
  else
    {
    timer1_period = nvd.timer1_period;
    baud_waits = nvd.baud_waits;
    ints_second = nvd.ints_second;
    freq_scale = 2;
    }
  }
  


void setup()
  {
  init_controller();
  pinMode(6,INPUT);         // these 2 pins are used with the analog comparator. high Z them for now
  pinMode(7,INPUT);
  
  ACSR  |=  (1<<ACI);    // clear Analog Comparator interrupt
  ACSR  |=
        (0<<ACD)   |         // Comparator ON
        (0<<ACBG)  |         // Disconnect 1.23V reference from AIN0 (use AIN0 and AIN1 pins)
        (0<<ACIE)  |         // Comparator Interrupt enabled
        (0<<ACIC)  |         // input capture disabled
        (0<<ACIS1) |         // set interrupt bit on rising edge
        (0<<ACIS0);          //    (ACIS1 and ACIS0 == 11)

  pinMode(2,INPUT_PULLUP);    // used to limit current via an external digital control. pull down to limit current. Let float to have no effect
  pinMode(3,INPUT_PULLUP);    // EQ push button control. press to start. press again to stop.

  pinMode(8,OUTPUT);          // for firmware development, a pin to drive to give me information
  pinMode(18,OUTPUT);         // my software serial port, 9600 baud, uses pin A4
  pinMode(19,OUTPUT);         // fan control, uses pin A5
  
  // timer1 20kHz with variable duty cycle
  // uses pin 9 (D9) as PWM output
  // PWM duty is set with OCR1A register
  // PWM freq is set with ICR1 resiter
  // units are cpu clock/2 i.e. 8MHz
  TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 1;
  ICR1 = timer1_period;         // gives 19,200 Hz actually. I use this as 2x serial port speed of 9600 Baud
  TIMSK1 |= (1 << TOIE1);   // enable timer overflow interrupt
  TIMSK0 = 0;
  
 
  Serial.begin(115200);
  get_biquad(0.1,2,0);    // LP filter setup
  get_biquad(0.1,4,1);    // even tho the inputs arrive after analog LP filtering, I like to do a bit more    
  get_biquad(0.1,4,2);    // the digital filter also adds a time delay which the code has been
  get_biquad(0.1,4,3);    // optimised for.
  
  cc = 0;                 // set up the local 80 char LCD buffer. This permits rapid updating of the info
  bcount = 0;             // with a slow LCD data transfer, so as to not upset the converter's pwm performance
  bp = &buf[0];           // 
  one_min = 50;           // force a mppt scan and potential tracking mode soon after boot (10 secs)
 }

void dlay(int n)
{
  int d;
  if (freq_scale == 2)
    d = n * 8;
  else
    d = n;
  dlay_timer = 0;         // a simple delay loop that uses the 20kHz pwm interrupt as a clock 
  while(dlay_timer < d) ; // This while loop will only exit if pwm interrupt is working.
}

ISR(TIMER1_OVF_vect)        
{  
// the main and only interrupt code 
// check for over current, send a character out the designated serial port pin, update pwm duty width
// increment the delay counter and maintain second, minute and hour counters
//
// over current check. Uses comparator. Fast. 
// If using a battery as the input, high current will momentarilly be present
// during the mppt scan. Even small increases in pwm duty % result in huge increase
// in power with a battery as input and low resistiance load.
if ((ACSR & 0x10) == 0x10 && nvd.fast_oc_enable != 0)
  {       
    OCR1A = 1;
    pwm = 1;
    erno = 6; 
    ACSR = 0; 
  }

OCR1A = pwm;    // pin 9 will be driven by this PWM duty cycle. Freq = 20kHz
one_sec++;
dlay_timer++;   // unsigned 16 bit integer counter for delay functions

// send one char out pin A4. First, get a character and put it in "blcd" and init state machine.
if (go_lcd == 0 && baudc > 50)    // not send as soon as possible, let the LCD digest the data with a delay here
  {
  // get next char
  blcd = buf[bcount];
  bcount++;
  if (bcount > 79)
    {
    bcount = 0;
    blcd =  '$';     // this is translated by the serial LCD as "put cursor at top left position"
    }   
  go_lcd = 1;
  bpos = 0;
  baudc = 2;
  }
  
if (go_lcd == 1 && baudc > baud_waits)   // a little 8 bit serial output state machine
  {
    if(bpos == 9)
    {
    // send stop bit, pull high
    sbi(PORTC,4);
    go_lcd = 0;
    baudc = 0;
    bpos = 0;
    }
  if (bpos > 0  )
    {
      if (blcd & 1)
        sbi(PORTC,4);
      else
        cbi(PORTC,4);
      blcd = blcd >> 1;
      bpos++;  
      baudc = 0;
    }
  if(bpos == 0 && go_lcd == 1)
    {
    // send start bit, pull low
    cbi(PORTC,4);
    bpos++;  
    baudc = 0;
    }
  }
  baudc++;              // a counter, goes from 0, 1, 0 , 1 etc. to halve the output frequency of the serial data

// now look after the time flags.

if (one_sec > ints_second)    // obtain a flag that is set once per second. My Nano has a crystal that needs this
  {                     // for 1 second
    one_sec = 0;
    one_sec_flag = 1;
    one_min++;

    if (one_min > 59)   // and a flag set once per minute and once an hour
      {
        one_min = 0;
        one_min_flag = 1;
        hour_count++;
        if (hour_count > 59)    // the max time needed to service this int is 6uS
          {                     // and the period is 50uS @ 20kHz          
            hour_count = 0;
            hour_flag = 1;
          }
      }
  } 
}   


float get_temp(int Vo,int hi_low)
{
 float a = 1.027957148e-3, b = 2.393569459e-4, c = 1.555071887e-7;
 float T,logRt,Tc,v,r,i;                                 // assumes NTC on high side of voltage divider
 v = 5.0 * (float)Vo /1024.0;    // v in volts
 if (hi_low == 1)     // NTC is on high side of divider, with 10K on low side
  {
    i = v/1e4;
    r = (5.0 - v)/i;
  }
 else                 // NTC is on low side of divider with 10K on high side
  {
    i = (5.0 - v)/1e4;
    r = v/i;
  }
 logRt = log(r); 
 T = (1.0 / (a + b*logRt + c*logRt*logRt*logRt)); // We get the temperature value in Kelvin from this Stein-Hart equation
 Tc = T - 273.15;                                 // Convert Kelvin to Celsius
 return Tc;
}



void get_data()
  {     
  float fa0,fa1,fa2,fa3; 
  
   fa0 = (float)analogRead(VOUT);  // get output voltage   
   biquad_filter(&fa0,0); 
   vout = fa0 * nvd.vout_cal;   // use the actual output voltage
   //vout = 12.0 + ((float)analogRead(5) - 512.0)/50.0;     // or a pot, for testing the charge state changes.
   
   fa1 = (float)analogRead(VIN);  // get input voltage
   biquad_filter(&fa1,1); 
   vin = fa1 * nvd.vin_cal;       // units are Volts
   
   fa2 = (float)(analogRead(IOUT) - nvd.z_iout);  // get output current. bipolar sensor so subtract zero offset
   biquad_filter(&fa2,2); 
   iout = fa2 * nvd.iout_cal;     // units are Amps
   
   fa3 = (float)(analogRead(IIN) - nvd.z_iin);  // get input current. bipolar sensor so subtract zero offset
   //biquad_filter(&fa3,3); 
   iin = fa3 * nvd.iin_cal; 
   
   ipwr = iin * vin;            // get input power and output power, in Watts
   opwr = iout * vout;
   
   hs_temp = get_temp(analogRead(HEATSINK),HI_LOW);
   inductor_temp = get_temp(analogRead(INDUCTOR),HI_LOW);
   
   // testing
   //hs_temp = 20.0;
   //inductor_temp = 22.0;

   // heatsink output current throttle
   // this will produce a modifed maximum current for tracking to follow.
   // as the temp rises from hs_thorttle_low to hs_max
   // the output current limit will be reduced from nvd.imax to zero
   //
   hs_throttle_current = nvd.imax;
   hs_throttle_flag = 0;
   if (hs_temp > nvd.hs_throttle_low && hs_temp < nvd.hs_max)
    {
      // only if the above limits are true do we throttle it
      hs_throttle_current = nvd.imax * (nvd.hs_max - hs_temp) / (nvd.hs_max - nvd.hs_throttle_low);
      hs_throttle_flag = 1;
    }
    
   ok_temps = 1;            // by default, unless the following tests prove otherwise
   if (hs_temp > nvd.fan_on)  
    {
      digitalWrite(19,1);     // heatsink fan control
      fan_on_flag = 1;
    }
   if (hs_temp < nvd.fan_off)  
    {
      digitalWrite(19,0);
      fan_on_flag = 0;
    }
   if (hs_temp > nvd.hs_max || inductor_temp > nvd.inductor_max)
    {
      buck_stop(4);
      ok_temps = 0;
      erno = 7;
    }  
   // low pass filter versions of the above, for humans (and me) to read from the LCD
   slow_iout = slow_iout * 0.9 + 0.1 * iout;
   slow_vout = slow_vout * 0.9 + 0.1 * vout;
   slow_iin = slow_iin * 0.9 + 0.1 * iin;
   slow_vin = slow_vin * 0.9 + 0.1 * vin;
  }

void do_data()
  {
  float p,e;
  char fbuf[10],hbuf[15],gbuf[10];
  int  i,q;
  if (pause_data == 1)
    {
      p = vin; if (p <0.0) p = 0; if (p >100.0) p = 100.0;      Serial.print(p);Serial.print(" ");
      p = vout; if (p <0.0) p = 0; if (p >100.0) p = 100.0;      Serial.print(p);Serial.print(" ");
      p = iin; if (p <0.0) p = 0; if (p >100.0) p = 100.0;      Serial.print(p);Serial.print(" ");
      p = iout; if (p <0.0) p = 0; if (p >100.0) p = 100.0;      Serial.print(p);Serial.print(" ");
      p = hs_temp; if (p <0.0) p = 0; if (p >100.0) p = 100.0;      Serial.print(p);Serial.print(" ");
      Serial.println();
    }
      // write controller run info into a 20 x 4 char buffer for an LCD 
      // this a huge bit of code and will be happily interrupted by the TIMER1 interrupt as needed.
      // 
      dtostrf(slow_vin,4,1,fbuf);  
      dtostrf(slow_iin,4,1,hbuf);  
      dtostrf(hs_temp,3,0,gbuf);     
      sprintf(bp,"In  %4sV %4sA %3sC",fbuf,hbuf,gbuf); 
      dtostrf(slow_vout,4,1,fbuf);  
      dtostrf(slow_iout,4,1,hbuf);   
      dtostrf(inductor_temp,3,0,gbuf);      
      sprintf(bp+40,"Out %4sV %4sA %3sC",fbuf,hbuf,gbuf);  
      if (hs_throttle_current < nvd.imax)
        buf[19] = '!';
      if ((iout > nvd.imax) || (digitalRead(2) == 0))
        buf[55] = '*';
      if (vout > (bv_target - 0.2))
        buf[49] = '*';  
      if (fan_on_flag == 1) 
        buf[16] = '*'; 
      p = slow_vout * slow_iout;
      if (p < 0.0) p = 0.0;
      dtostrf(p,5,0,fbuf);   
      e = (hours_energy + nvt.total_energy);
      if (e < 0.0) e = 0.0;
      dtostrf(e,6,1,hbuf);        
      sprintf(bp+20,"%5sW Out %6skW",fbuf,hbuf); buf[39] = 'h';
      
      if (track_mode == 0) strcpy_P(fbuf,(char *)F("NIGHT "));
      if (track_mode == 1) strcpy_P(fbuf,(char *)F("MPPT  "));
      if (track_mode == 2) strcpy_P(fbuf,(char *)F("ABSORB"));
      if (track_mode == 3) strcpy_P(fbuf,(char *)F("FLOAT "));
      if (track_mode == 4) strcpy_P(fbuf,(char *)F("EQ    "));
      if (calibrate == 1)  strcpy_P(fbuf,(char *)F("CAL   "));
      if (erno == 0) strcpy_P(hbuf,(char *)F("Vin < Bv  !"));
      if (erno == 1) strcpy_P(hbuf,(char *)F("Bv > BvMax!"));
      if (erno == 2) strcpy_P(hbuf,(char *)F("EQ Cancel  "));
      if (erno == 3) strcpy_P(hbuf,(char *)F("STOPPED    "));  
      if (erno == 4 || erno == 3) 
        {
          dtostrf(bv_target,3,1,gbuf);
          sprintf(hbuf," %3sV %2d  ",gbuf,one_min);
        }
      if (erno == 5) strcpy_P(hbuf,(char *)F("DEAD       "));
      if (erno == 6) strcpy_P(hbuf,(char *)F("OVERCURRENT"));
      if (erno == 7) strcpy_P(hbuf,(char *)F("OVERTEMP"));
      
      sprintf(bp+60,"%5s%s",fbuf,hbuf); 
      if (is_running == 1 && track_mode != MPPT)
        {
        if (track_mode == ABSORB)
          i = (int)absorb_time;
        if (track_mode == EQ)
          i = (int)eq_time;  
        if (track_mode == FLOAT)
          i = (int)float_time;    
        q = i / 100;
        buf[77] = '0' + q;  
        q = i % 10;
        buf[79] = '0' + q;
        q = (i / 10) % 10;
        buf[78] = '0' + q;
        }
      else
        {
          buf[77] = buf[78] = buf[79] = '-';
        }
  }

void mppt_scan()
  {
   //  find the pwm duty cycle width that gives the highest input power measurement
   // search up to near full pwm duty %, battery target voltage or max output current, which ever occurs first.
   
  int i,di;
  float  siout,svout,pwr,oldpwr;
  int start_pwm;
  // maybe buck converter is stopped, PWM output would be disabled
  // so enable it if needed. DO not enable it when running, PWM will be immediately taken to 1
  // and likely ruin the tracking
  if (is_running == 0)
    buck_start();
  // get start point for search, set max found to zero
  max_pwr = 0.0;
  start_pwm = pwm;
  
  // smoothly ramp down to small pwm from start pwm. 5 cycle pwm width of the 800 cycle wide full power
  // is not a lot. I call it very low power.
  di = 20;
  for(i = pwm; i > (di+5); i -= di)
    {
    pwm = i;
    dlay(100); 
    }

  // search for max power, and record pwm for that power, from something low to nearly full power 
  // search with an adaptive increment.
  // when the change in power is small, increase pwm fast, 10 counts per loop
  // but once power change is large, increase by 1 count.
  // this avoids over shoot into over current trigger territory.
  
  di = 20 ;        // step size. I needed to change this when output power increases too fast, when a battery is used for input.
  oldpwr = 0.0;
  
  for(i = 5; i < (timer1_period - di); i += di)
    {
    pwm = i;      // very shortly after this, the pwm width will be set with this new value
                  // Let things stabilise a bit. There are large capacitors on the input and output of the converter
                  // so there will be a time delay and it will be different with all the different builds.
    dlay(100); 
    
    // get the dc-dc converter vital signs
    siout = (float)(analogRead(IOUT) - nvd.z_iout) * nvd.iout_cal;  
    svout = (float)analogRead(VOUT) * nvd.vout_cal;
    // get output power
    pwr =  siout * svout;
    
    oldpwr = pwr;
    // do search
    if (pwr > max_pwr) 
      {
        max_pwr = pwr;
        max_pwr_pwm = i;
      }
     // do not exceed max current setting, nor battery target voltage. if so break out of this loop
    if (siout > (nvd.imax * 1.05) || svout > (bv_target*1.05) )
      break;  
    }
    
  // ramp back down to start position, down from the search end point down to start_pwm
  di = 20 ;
  for(; i > (start_pwm + di); i -= di)
    {
    pwm = i;  
    dlay(100) ; 
    } 
  if (max_pwr > nvd.night_pwr)
    track_mode = MPPT;  
  }

void loop()
  { 
  int i,di;

  // deal with user input via USB serial port    
  get_data(); 
  if (got_data == 1)
    parse_it();
  
  // update controller information
  if (one_sec_flag == 1)
    {
    one_sec_flag = 0;
    bv_target = get_bv_track_mode();    // set battery target, dependent on the mode (just get to absorb, float or EQ)
    sanity_check();  
    do_data();  
    }
    
  if (is_running == 1 && calibrate == 0)    
    do_tracking();
  
  if (one_min_flag == 1 && calibrate == 0)   
    {   // once a minute do a power scan. The solar array might not have one peak, could have 2. maybe.
      one_min_flag = 0;
      if (is_running == 1)
        hours_energy = hours_energy + 1e-3*(slow_vout * slow_iout) / SEC_MIN;      // in units of k Watt.hours
      // once an hour update energy total
      if (hour_flag == 1)
        {
          hour_flag = 0;                 // can get total energy from (hours_energy + ndt.total_energy)
          if (is_running == 1)
          {
            EEPROM.get(2,nvt);             // EEPROM wear capacity is ample for read/write cycle of 1 update/hour
            nvt.total_energy += hours_energy; // 100,000 write cycles
            EEPROM.put(2,nvt);
          }
          hours_energy = 0.0;
        }                 
      if(  (track_mode == MPPT || track_mode == NIGHT) && hs_throttle_flag == 0) 
        mppt_scan();  
      if ( is_running == 1 )
        {
        // ramp to located max power pwm and it will stay there
        if(track_mode == MPPT)
          {
          di = 10 / freq_scale;
          if (pwm < max_pwr_pwm)    // but can't assume the change is in one or the other direction
            for(i=pwm; i <= max_pwr_pwm ; i += di)
                {
                pwm = i;
                dlay(50);   
                }

          else
            for(i=pwm; i >= max_pwr_pwm ; i -= di)
              {
                pwm = i;
                dlay(50);   
              } 
          }
        }
    }         // one_min flag  
  }

void do_tracking()
  {
    int nupwm,d2, dec_val, inc_val;
    dd++;
    if (dd < 20) return;  
    dd=0;
    //
    // we can have an extra "current limit" via this input. When high it does nothing.
    // when pulled down to ground, it will be a current limit, in effect making output current ALWAYS exceed
    // max current set point.
    // Mark will need this with his parallel controllers.
    //
    d2 = digitalRead(2);
      
    // make copy of current pwm so changes made to it will not be fed into the timer immediately.
    // we need to do some limit checks before the timer can have it  
    nupwm = pwm;
 
    dec_val = 4 / freq_scale;    // these define how fast pwm is altered. I set these from trial and error
    inc_val = 2 / freq_scale;
    // take the fact that when pwm = 0, then zero power out. If increase pwm then power out increases
    // But only if we assume pwm will never exceed max_pwr_pwm. This means the power -> pwm curve is near zero at pwm = 0
    // and max power at pwm = max_pwr_pwm. Permit pwm changes to occur only within these bounds. 
    if ((vout > bv_target) || (iout > hs_throttle_current) || (d2 == 0) || (vin < bv_target))   nupwm -= dec_val;  // protect controller and battery
    if ((vout < bv_target) && (iout < hs_throttle_current) && (d2 == 1) && (vin > bv_target))   nupwm += inc_val;  // but try to get back to mp
    if (nupwm > max_pwr_pwm) nupwm = max_pwr_pwm;        
    // now clamp to reasonable values no matter what and then let timer use the new value
    if (nupwm > (timer1_period - 1)) nupwm = timer1_period - 1;
    if (nupwm < 1) nupwm = 1;
    pwm = nupwm;
  }

void sanity_check()
  {
  // check if battery connected, solar input exists, determine tracking mode, etc.
  // this is called once a second
  //

  if (timeout_check(5,(erno ==6)))      // OC was triggered. The pwm is dead, controller needs a reboot and will after this timeout
    reset_uc();
  
  if (abs(vin - vout) < 1.0)      // buck converters always have battery voltage or more present on input
    {                             // thanks to the MOSFET's body diode
    track_mode = NIGHT;           // night time check
    absorb_time = 0.0;
    buck_stop(3);
    }
  
  if  ( timeout_check(4, (vout < (bv_target-nvd.absorb_tol)) && track_mode == ABSORB ))
    {
    track_mode = MPPT;          // bv dropped below some limit? no more absorb, now it's mppt
    absorb_time = 0.0;          // always zero absorb time total, ready for transtion from mppt to absorb
    }


  if (track_mode == ABSORB)       // count time in absorb
    absorb_time += 1.0/SEC_MIN;
        
  if ( (vout > (bv_target-nvd.absorb_tol)) && (is_running == 1) && track_mode != FLOAT && track_mode != EQ )
    {
    track_mode = ABSORB;        // this looks for the transition from NIGHT or MPPT
    float_time = 0.0;           // to ABSORB
    }

  if (track_mode == FLOAT)        // count time in FLOAT
    float_time += 1.0/SEC_MIN;
  
 if (absorb_time > nvd.absorb_float_time && track_mode == ABSORB)
    {
    // go to float
    track_mode = FLOAT;   // we float for ever if enough solar incoming. 
    absorb_time = 0.0;    // reset for when bv drops, causing a new period of absorb
    float_time = 0.0;
    }
  
  if (track_mode == EQ)
    {
    eq_time += 1.0/SEC_MIN;
    if (eq_time > nvd.eq_duration)
      {
      // back to MPPT or ABSORB, depending on output voltage
      track_mode = MPPT;
      absorb_time = 0.0;   
      float_time = 0.0;
      erno = 4;
      }
    }
    
  if ( timeout_check(0,(eq_button == 1) && (vout < nvd.eq_reset) ) )
    {
    eq_button = 0;            // if bv drops below some limit for EQ, then cancel it
    track_mode = MPPT;
    erno = 4;
    }
  
  if (timeout_check(1, (vout > nvd.vbat_max) ) )
    {
    buck_stop(1);              // save battery from over charge if tracking fails
    erno = 1;
    }  

  if ( timeout_check(3,(vout < (bv_target - nvd.float_tol) && track_mode == FLOAT) ) )
    {
    // cancel float, go back to absorb/mppt
    float_time = 0.0;
    track_mode = MPPT; 
    }
  
  if ( (digitalRead(3) == 0))    // EQ button pressed.
    {
    if(eq_button == 0)
      {
      eq_button = 1;
      // start EQ, set EQ variables to start state
      track_mode = EQ;
      eq_time = 0.0;
      }
    else
      {     
      eq_button = 0;
      // cancel EQ, reset EQ stuff to intitial state
      eq_time = 0.0;
      track_mode == ABSORB;
      }      
    }
  }
